import asyncio

import pytest

from kopf._kits.webhooks import WebhookNgrokTunnel


async def test_missing_pyngrok(no_pyngrok, responder):
    with pytest.raises(ImportError) as err:
        server = WebhookNgrokTunnel()
        async with server:
            async for _ in server(responder.fn):
                break  # do not sleep
    assert "pip install pyngrok" in str(err.value)


async def test_ngrok_tunnel(
        certfile, pkeyfile, responder, pyngrok_mock):

    responder.fut.set_result({'hello': 'world'})
    server = WebhookNgrokTunnel(port=54321, path='/p1/p2',
                                region='xx', token='xyz', binary='/bin/ngrok')
    async with server:
        async for client_config in server(responder.fn):
            assert 'caBundle' not in client_config  # trust the default CA
            assert client_config['url'] == 'https://nowhere/p1/p2'
            break  # do not sleep

    assert pyngrok_mock.conf.get_default.called
    assert pyngrok_mock.conf.get_default.return_value.ngrok_path == '/bin/ngrok'
    assert pyngrok_mock.conf.get_default.return_value.region == 'xx'
    assert pyngrok_mock.ngrok.set_auth_token.called
    assert pyngrok_mock.ngrok.set_auth_token.call_args_list[0].args[0] == 'xyz'
    assert pyngrok_mock.ngrok.connect.called
    assert pyngrok_mock.ngrok.connect.call_args_list[0].args[0] == '54321'
    assert pyngrok_mock.ngrok.connect.call_args_list[0].kwargs['bind_tls'] == True
    assert pyngrok_mock.ngrok.disconnect.called

    await asyncio.get_running_loop().shutdown_asyncgens()



import asyncio
import textwrap

import aiohttp.web
import pytest

from kopf._cogs.clients.api import delete, get, patch, post, request, stream
from kopf._cogs.clients.errors import APIError

pytestmark = pytest.mark.usefixtures('fake_vault')


@pytest.mark.parametrize('method', ['get', 'post', 'patch', 'delete'])
async def test_raw_requests_work(kmock, method, settings, logger):
    api = kmock[method, '/url'] << {}

    response = await request(
        method=method,
        url='/url',
        payload={'fake': 'payload'},
        headers={'fake': 'headers'},
        settings=settings,
        logger=logger,
    )
    assert isinstance(response, aiohttp.ClientResponse)  # unparsed!
    assert len(api) == 1
    assert api[0].method.lower() == method
    assert api[0].url.path == '/url'
    assert api[0].data == {'fake': 'payload'}
    assert api[0].headers['fake'] == 'headers'  # and other system headers


@pytest.mark.parametrize('method', ['get', 'post', 'patch', 'delete'])
async def test_raw_requests_are_not_parsed(kmock, method, settings, logger):
    kmock[method, '/url'] << b'BAD JSON!'
    response = await request(method, '/url', settings=settings, logger=logger)
    assert isinstance(response, aiohttp.ClientResponse)


@pytest.mark.parametrize('method', ['get', 'post', 'patch', 'delete'])
async def test_server_errors_escalate(kmock, method, settings, logger):
    kmock[method, '/url'] << {} << 666
    with pytest.raises(APIError) as err:
        await request(method, '/url', settings=settings, logger=logger)
    assert err.value.status == 666


@pytest.mark.parametrize('method', ['get', 'post', 'patch', 'delete'])
async def test_relative_urls_are_prepended_with_server(kmock, fake_vault, method, settings, logger):
    kmock[method, '/url'] << {}
    await request(method, '/url', settings=settings, logger=logger)
    assert len(kmock) == 1
    assert kmock[0].url.host == kmock.url.host


@pytest.mark.parametrize('method', ['get', 'post', 'patch', 'delete'])
@pytest.mark.kmock(hostnames=['fakehost.fakedomain.tld'])
async def test_absolute_urls_are_passed_through(kmock, fake_vault, method, settings, logger):
    kmock[method, '/url'] << {}
    await request(method, 'http://fakehost.fakedomain.tld/url', settings=settings, logger=logger)
    assert len(kmock) == 1
    assert kmock[0].url.host == 'fakehost.fakedomain.tld'


@pytest.mark.parametrize('fn, method', [
    (get, 'get'),
    (post, 'post'),
    (patch, 'patch'),
    (delete, 'delete'),
])
async def test_parsing_in_requests(kmock, fn, method, settings, logger):
    api = kmock[method, '/url'] << {'fake': 'result'}
    response = await fn(
        url='/url',
        payload={'fake': 'payload'},
        headers={'fake': 'headers'},
        settings=settings,
        logger=logger,
    )
    assert response == {'fake': 'result'}  # parsed!
    assert len(api) == 1
    assert api[0].method.lower() == method
    assert api[0].url.path == '/url'
    assert api[0].data == {'fake': 'payload'}
    assert api[0].headers['fake'] == 'headers'  # and other system headers


@pytest.mark.parametrize('method', ['get'])  # the only supported method at the moment
async def test_parsing_in_streams(kmock, method, settings, logger):
    kmock[method, '/url'] << {"fake": "result1"} << {"fake": "result2"}

    items = []
    async for item in stream(
        url='/url',
        payload={'fake': 'payload'},
        headers={'fake': 'headers'},
        settings=settings,
        logger=logger,
    ):
        items.append(item)

    assert items == [{'fake': 'result1'}, {'fake': 'result2'}]
    assert len(kmock) == 1
    assert kmock[0].method.lower() == method
    assert kmock[0].url.path == '/url'
    assert kmock[0].data == {'fake': 'payload'}
    assert kmock[0].headers['fake'] == 'headers'  # and other system headers


@pytest.mark.parametrize('fn, method', [
    (get, 'get'),
    (post, 'post'),
    (patch, 'patch'),
    (delete, 'delete'),
])
async def test_direct_timeout_in_requests(kmock, fn, method, settings, logger, looptime):
    kmock[method, '/url'] << (lambda: asyncio.sleep(10)) << {}

    with pytest.raises(asyncio.TimeoutError):
        timeout = aiohttp.ClientTimeout(total=1.23)
        # aiohttp raises an asyncio.TimeoutError which is automatically retried.
        # To reduce the test duration we disable retries for this test.
        settings.networking.error_backoffs = None
        await fn('/url', timeout=timeout, settings=settings, logger=logger)

    assert looptime == 1.23


@pytest.mark.parametrize('fn, method', [
    (get, 'get'),
    (post, 'post'),
    (patch, 'patch'),
    (delete, 'delete'),
])
async def test_settings_timeout_in_requests(kmock, fn, method, settings, logger, looptime):
    kmock[method, '/url'] << (lambda: asyncio.sleep(10)) << {}

    with pytest.raises(asyncio.TimeoutError):
        settings.networking.request_timeout = 1.23
        # aiohttp raises an asyncio.TimeoutError which is automatically retried.
        # To reduce the test duration we disable retries for this test.
        settings.networking.error_backoffs = None
        await fn('/url', settings=settings, logger=logger)

    assert looptime == 1.23


@pytest.mark.parametrize('method', ['get'])  # the only supported method at the moment
async def test_direct_timeout_in_streams(kmock, method, settings, logger, looptime):

    kmock[method, '/url'] << (lambda: asyncio.sleep(10)) << {}

    with pytest.raises(asyncio.TimeoutError):
        timeout = aiohttp.ClientTimeout(total=1.23)
        # aiohttp raises an asyncio.TimeoutError which is automatically retried.
        # To reduce the test duration we disable retries for this test.
        settings.networking.error_backoffs = None
        async for _ in stream('/url', timeout=timeout, settings=settings, logger=logger):
            pass

    assert looptime == 1.23


@pytest.mark.parametrize('method', ['get'])  # the only supported method at the moment
async def test_settings_timeout_in_streams(kmock, method, settings, logger, looptime):
    kmock[method, '/url'] << (lambda: asyncio.sleep(10)) << {}

    with pytest.raises(asyncio.TimeoutError):
        settings.networking.request_timeout = 1.23
        # aiohttp raises an asyncio.TimeoutError which is automatically retried.
        # To reduce the test duration we disable retries for this test.
        settings.networking.error_backoffs = None
        async for _ in stream('/url', settings=settings, logger=logger):
            pass

    assert looptime == 1.23


@pytest.mark.parametrize('delay, expected_times, expected_items', [
    pytest.param(0, [], [], id='instant-none'),
    pytest.param(2, [1], [{'fake': 'result1'}], id='fast-single'),
    pytest.param(9, [1, 4], [{'fake': 'result1'}, {'fake': 'result2'}], id='inf-double'),
])
@pytest.mark.parametrize('initial', [None, b''], ids=['slow-headers', 'fast-headers'])
@pytest.mark.parametrize('method', ['get'])  # the only supported method at the moment
async def test_stopper_in_streams(
        kmock, method, delay, initial, settings, logger, looptime,
        expected_items, expected_times):
    kmock[method, '/url'] << (
        initial,  # prepare and send the headers before the sleep
        lambda: asyncio.sleep(1),
        {"fake": "result1"},
        lambda: asyncio.sleep(3),
        {"fake": "result2"},
    )

    stopper = asyncio.Future()
    asyncio.get_running_loop().call_later(delay, stopper.set_result, None)

    items = []
    times = []
    async for item in stream('/url', stopper=stopper, settings=settings, logger=logger):
        items.append(item)
        times.append(float(looptime))

    assert items == expected_items
    assert times == expected_times



import asyncio

import pytest

from kopf._cogs.aiokits.aiovalues import Container


async def test_empty_by_default(looptime):
    container = Container()
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(container.wait(), timeout=9)
    assert looptime == 9


async def test_does_not_wake_up_when_reset(looptime):
    container = Container()

    async def reset_it():
        await container.reset()

    loop = asyncio.get_running_loop()
    loop.call_later(1, asyncio.create_task, reset_it())

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(container.wait(), timeout=9)
    assert looptime == 9


async def test_wakes_up_when_preset(looptime):
    container = Container()
    await container.set(123)

    result = await container.wait()
    assert looptime == 0
    assert result == 123


async def test_wakes_up_when_set(looptime):
    container = Container()

    async def set_it():
        await container.set(123)

    loop = asyncio.get_running_loop()
    loop.call_later(9, asyncio.create_task, set_it())

    result = await container.wait()
    assert looptime == 9
    assert result == 123


async def test_iterates_when_set(looptime):
    container = Container()

    async def set_it(v):
        await container.set(v)

    loop = asyncio.get_running_loop()
    loop.call_later(6, asyncio.create_task, set_it(123))
    loop.call_later(9, asyncio.create_task, set_it(234))

    values = []
    async for value in container.as_changed():
        values.append(value)
        if value == 234:
            break

    assert looptime == 9
    assert values == [123, 234]


async def test_iterates_when_preset(looptime):
    container = Container()
    await container.set(123)

    values = []
    async for value in container.as_changed():
        values.append(value)
        break

    assert looptime == 0
    assert values == [123]



import asyncio
import dataclasses
import functools
import os
from collections.abc import Callable, Collection
from typing import Any

import click

from kopf._cogs.aiokits import aioadapters
from kopf._cogs.configs import configuration
from kopf._cogs.helpers import loaders
from kopf._cogs.structs import credentials, references
from kopf._core.actions import loggers
from kopf._core.engines import peering
from kopf._core.intents import registries
from kopf._core.reactor import running
from kopf._kits import loops


@dataclasses.dataclass()
class CLIControls:
    """ :class:`KopfRunner` controls, which are impossible to pass via CLI. """
    ready_flag: aioadapters.Flag | None = None
    stop_flag: aioadapters.Flag | None = None
    vault: credentials.Vault | None = None
    registry: registries.OperatorRegistry | None = None
    settings: configuration.OperatorSettings | None = None
    loop: asyncio.AbstractEventLoop | None = None


def logging_options(fn: Callable[..., Any]) -> Callable[..., Any]:
    """ A decorator to configure logging in all commands the same way."""
    @click.option('-v', '--verbose', is_flag=True)
    @click.option('-d', '--debug', is_flag=True)
    @click.option('-q', '--quiet', is_flag=True)
    @click.option('--log-format', type=click.Choice(loggers.LogFormat, case_sensitive=False), default='full')
    @click.option('--log-refkey', type=str)
    @click.option('--log-prefix/--no-log-prefix', default=None)
    @functools.wraps(fn)  # to preserve other opts/args
    def wrapper(verbose: bool, quiet: bool, debug: bool,
                log_format: loggers.LogFormat = loggers.LogFormat.FULL,
                log_prefix: bool | None = False,
                log_refkey: str | None = None,
                *args: Any, **kwargs: Any) -> Any:
        loggers.configure(debug=debug, verbose=verbose, quiet=quiet,
                          log_format=log_format, log_refkey=log_refkey, log_prefix=log_prefix)
        return fn(*args, **kwargs)

    return wrapper


@click.group(name='kopf', context_settings=dict(
    auto_envvar_prefix='KOPF',
))
@click.version_option(prog_name='kopf')
@click.make_pass_decorator(CLIControls, ensure=True)
def main(__controls: CLIControls) -> None:
    pass


@main.command()
@logging_options
@click.option('-A', '--all-namespaces', 'clusterwide', is_flag=True)
@click.option('-n', '--namespace', 'namespaces', multiple=True)
@click.option('--standalone', is_flag=True, default=None)
@click.option('--dev', is_flag=True)
@click.option('-L', '--liveness', 'liveness_endpoint', type=str)
@click.option('-P', '--peering', 'peering_name', type=str, envvar='KOPF_RUN_PEERING')
@click.option('-p', '--priority', type=int)
@click.option('-m', '--module', 'modules', multiple=True)
@click.argument('paths', nargs=-1)
@click.make_pass_decorator(CLIControls, ensure=True)
def run(
        __controls: CLIControls,
        paths: list[str],
        modules: list[str],
        peering_name: str | None,
        priority: int | None,
        dev: bool | None,
        standalone: bool | None,
        namespaces: Collection[references.NamespacePattern],
        clusterwide: bool,
        liveness_endpoint: str | None,
) -> None:
    """ Start an operator process and handle all the requests. """
    priority = 666 if dev else priority
    if os.environ.get('KOPF_RUN_NAMESPACE'):  # legacy for single-namespace mode
        namespaces = tuple(namespaces) + (os.environ.get('KOPF_RUN_NAMESPACE', ''),)
    if namespaces and clusterwide:
        raise click.UsageError("Either --namespace or --all-namespaces can be used, not both.")
    if __controls.registry is not None:
        registries.set_default_registry(__controls.registry)
    loaders.preload(
        paths=paths,
        modules=modules,
    )
    with loops.proper_loop(suggested_loop=__controls.loop) as actual_loop:
        return running.run(
            standalone=standalone,
            namespaces=namespaces,
            clusterwide=clusterwide,
            priority=priority,
            peering_name=peering_name,
            liveness_endpoint=liveness_endpoint,
            registry=__controls.registry,
            settings=__controls.settings,
            stop_flag=__controls.stop_flag,
            ready_flag=__controls.ready_flag,
            vault=__controls.vault,
            loop=actual_loop,
        )


@main.command()
@logging_options
@click.option('-n', '--namespace', 'namespaces', multiple=True)
@click.option('-A', '--all-namespaces', 'clusterwide', is_flag=True)
@click.option('-i', '--id', type=str, default=None)
@click.option('--dev', is_flag=True)
@click.option('-P', '--peering', 'peering_name', required=True, envvar='KOPF_FREEZE_PEERING')
@click.option('-p', '--priority', type=int, default=100, required=True)
@click.option('-t', '--lifetime', type=int, required=True)
@click.option('-m', '--message', type=str)
@click.make_pass_decorator(CLIControls, ensure=True)
def freeze(
        __controls: CLIControls,
        id: str | None,
        message: str | None,
        lifetime: int,
        namespaces: Collection[references.NamespacePattern],
        clusterwide: bool,
        peering_name: str,
        priority: int,
        dev: bool,
) -> None:
    """ Pause the resource handling in the operator(s). """
    priority = 666 if dev else priority
    identity = peering.Identity(id) if id else peering.detect_own_id(manual=True)
    insights = references.Insights()
    settings = configuration.OperatorSettings()
    settings.peering.name = peering_name
    settings.peering.priority = priority
    with loops.proper_loop(suggested_loop=__controls.loop) as actual_loop:
        return running.run(
            clusterwide=clusterwide,
            namespaces=namespaces,
            insights=insights,
            identity=identity,
            settings=settings,
            loop=actual_loop,
            _command=peering.touch_command(
                insights=insights,
                identity=identity,
                settings=settings,
                lifetime=lifetime))


@main.command()
@logging_options
@click.option('-n', '--namespace', 'namespaces', multiple=True)
@click.option('-A', '--all-namespaces', 'clusterwide', is_flag=True)
@click.option('-i', '--id', type=str, default=None)
@click.option('-P', '--peering', 'peering_name', required=True, envvar='KOPF_RESUME_PEERING')
@click.make_pass_decorator(CLIControls, ensure=True)
def resume(
        __controls: CLIControls,
        id: str | None,
        namespaces: Collection[references.NamespacePattern],
        clusterwide: bool,
        peering_name: str,
) -> None:
    """ Resume the resource handling in the operator(s). """
    identity = peering.Identity(id) if id else peering.detect_own_id(manual=True)
    insights = references.Insights()
    settings = configuration.OperatorSettings()
    settings.peering.name = peering_name
    with loops.proper_loop(suggested_loop=__controls.loop) as actual_loop:
        return running.run(
            clusterwide=clusterwide,
            namespaces=namespaces,
            insights=insights,
            identity=identity,
            settings=settings,
            loop=actual_loop,
            _command=peering.touch_command(
                insights=insights,
                identity=identity,
                settings=settings,
                lifetime=0))



import pytest

from kopf._core.reactor.queueing import EOS, get_version


@pytest.mark.parametrize('raw_event, expected_version', [
    pytest.param(EOS.token, None, id='eos'),
    pytest.param({}, None, id='empty-event'),
    pytest.param({'object': {}}, None, id='empty-object'),
    pytest.param({'object': {'metadata': {}}}, None, id='empty-metadata'),
    pytest.param({'object': {'metadata': {'resourceVersion': '123abc'}}}, '123abc', id='123abc'),
])
def test_resource_version_detection(raw_event, expected_version):
    resource_version = get_version(raw_event)
    assert resource_version == expected_version
